import os
from typing import Any

import cloudpickle
import numpy as np
import pandas as pd
import scipy.sparse as sparse


class PredictProbaWrapper(object):

    def __init__(self, model: Any, transformer: Any):
        self.model = model
        self.transformer = transformer

    def predict(self, df: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            self.model.predict_proba(df), columns=self.model.classes_
        )

    def transform(self, pre_transform_df: pd.DataFrame) -> pd.DataFrame:
        transformed_data = self.transformer.transform(pre_transform_df)
        if isinstance(transformed_data, pd.DataFrame):
            return transformed_data
        try:
            post_transform_features = self.transformer.get_feature_names_out()
        except:
            raise AssertionError(
                "Expected transform object to output a DataFrame or implement the `get_feature_names_out()` function to retrieve post features!"
            )
        if isinstance(transformed_data, sparse.csr_matrix):
            # We convert 0 to None because when we densify a sparse matrix, the intrepretation of missing value
            # in the XGB model changes. Look at https://github.com/dmlc/xgboost/issues/1634
            return pd.DataFrame(
                transformed_data.toarray(), columns=post_transform_features
            ).replace({0: np.nan})
        else:
            return pd.DataFrame(
                transformed_data, columns=post_transform_features
            )

    def get_model(self):
        return self.model


def _load_model_from_local_file(path: str):
    parent_dir = os.path.dirname(path)
    with open(path, "rb") as f:
        loaded_model = cloudpickle.load(f)
    with open(os.path.join(parent_dir, "transformer.pkl"), "rb") as t:
        loaded_transformer = cloudpickle.load(t)
    return PredictProbaWrapper(loaded_model, loaded_transformer)


def _load_pyfunc(path: str):
    return _load_model_from_local_file(path)
